# Modules

## Available modules

{{- range $key, $categoryModules := .Modules }}

### {{ $key }}
{{ range $categoryModules }}
- {{ . }}
{{- end }}
{{- end }}

## Bootstrapping

When getting started with buildr, adding new modules might be a little bit time-consuming at first.
To speed up your workflow buildr allows you to bootstrap modules quite quickly:

```shell
# buildr new <category> <module> [name]
buildr new task script hello_world
```

This will output an empty HCL block containing all possible options of the specified module.
This way you've only to remove what you don't need and fill the empty values with your configuration.

To directly append it to one of your existing configs:

```shell
buildr new task script hello_world >> tasks.hcl
```

## Input mappings

Input mappings are a feature to:

* narrow the scope of a module 
* map outputs of modules in a way so that they're usable but not mixed up with the original source code

this way it's possible to apparently write outputs to the repository root, but it's actually stored in a subdirectory per module.
In the background buildr starts a FUSE driver to mimic a UnionFS like behavior of multiple layers on top of each other with a write-layer on top to redirect writes.

#### Example

The following example illustrates how to use input mappings:

```hcl
task "script" "input_mapping_sample" {
  inline = [
    "touch test.txt"
  ]

  input_mapping = {
    "${buildr.repo.root}" = "."
  }
}
```

`buildr.repo.root` is a built-in variable that points to the discover repository root (or manually set via CLI arguments).
The repository root is mapped to the working directory of the module and the inline script creates a new file.
After executing the task you will notice that the text file is not created at the repository root but in the *out* directory of the module.

## Running in a container

It's possible to run a module in a container.
To do so, you've to define at least the container image in which the module should be executed.

**Note:** currently **only** Linux containers are supported, apart from that there are no further restrictions.

#### Example - basic: `buf lint` your protobuf files

The following examples illustrates how you can run a task in a container:

```hcl
task "script" "buf_generate" {
  inline = [
    "buf lint"
  ]

  container {
    image = "docker.io/bufbuild/buf"
  }
}
```

### Container volumes

*buildr* supports multiple types of container volumes:

* volume mounts
* bind mounts
* tmpfs mounts

#### Example - caching: `go build` 

When running consecutive `go build`s, downloading all necessary modules all the time can become very annoying.
To avoid something like that, it's possible to use persist the Go module cache:

```hcl
task "script" "go_build" {
  inline = [
    "go build ."
  ]

  container {
    image = "docker.io/golang:1.20"

    volume_mount "/go/pkg/mod" {
      name = "go_cache"
    }
  }
}
```

#### Example - the other mounts

```hcl
task "script" "docker_hello_world" {
  inline = [
    "docker run --rm -i docker.io/alpine:edge echo 'hello world'"
  ]

  environment = {
    DOCKER_HOST = "unix:///var/run/docker.sock"
  }

  container {
    image = "docker.io/alpine/docker-with-buildx"

    tmpfs_mount "/tmp" {
      read_only = false
      size      = 1024
    }

    bind_mount "/var/run/docker.sock" {
      source    = "/var/run/docker.sock"
      read_only = true
    }
  }
}
```