package cmd

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log/slog"
	"os"
	"os/exec"
	"strings"
	"time"

	"github.com/docker/docker/client"

	"code.icb4dc0.de/buildr/buildr/internal/slices"

	"code.icb4dc0.de/buildr/buildr/internal/config"
	"code.icb4dc0.de/buildr/buildr/internal/plugins"

	"code.icb4dc0.de/buildr/buildr/internal/lazy"

	"code.icb4dc0.de/buildr/buildr/assets"

	"code.icb4dc0.de/buildr/buildr/internal/hcl"

	"code.icb4dc0.de/buildr/buildr/internal/profiling"
	"code.icb4dc0.de/buildr/buildr/modules/state"

	"code.icb4dc0.de/buildr/buildr/internal/containers"
	"code.icb4dc0.de/buildr/buildr/internal/errs"
	"code.icb4dc0.de/buildr/buildr/internal/execution"
	"code.icb4dc0.de/buildr/buildr/internal/execution/container"
	"code.icb4dc0.de/buildr/buildr/internal/execution/local"
	"code.icb4dc0.de/buildr/buildr/internal/ignore"
	"code.icb4dc0.de/buildr/buildr/internal/logging"
	"code.icb4dc0.de/buildr/buildr/internal/vault"
	"code.icb4dc0.de/buildr/buildr/modules"
	"code.icb4dc0.de/buildr/buildr/modules/packaging"
	"code.icb4dc0.de/buildr/buildr/modules/repo"
	"code.icb4dc0.de/buildr/buildr/modules/task"

	hcl2 "github.com/hashicorp/hcl/v2"
	"github.com/hashicorp/hcl/v2/gohcl"
	"github.com/spf13/cobra"
)

var (
	cwd                         string
	ErrStateStoreNotInitialized = errors.New("state DB not initialized")
)

func init() {
	if wd, err := os.Getwd(); err != nil {
		panic(err)
	} else {
		cwd = wd
	}
}

func NewApp(ctx context.Context) *App {
	app := &App{
		rootCmd: &cobra.Command{
			Use:           "buildr",
			SilenceUsage:  true,
			SilenceErrors: true,
		},
		services: appServices{
			registry: modules.NewRegistry(
				task.Registration,
				packaging.Registration,
			),
			dockerClient: lazy.NewLazyCtx(ctx, prepareDockerAPIClient),
		},
		pluginMgr: &plugins.Manager{
			Downloader: plugins.GenericDownloader{
				"http":  plugins.HTTPDownloader{},
				"https": plugins.HTTPDownloader{},
				"file":  plugins.FileDownloader{},
			},
		},
		loggingCfg: logging.NewConfig(),
	}

	app.rootCmd.PersistentPreRunE = func(cmd *cobra.Command, _ []string) error {
		return app.Init(cmd.Context())
	}

	app.rootCmd.PersistentPostRunE = func(*cobra.Command, []string) error {
		return app.Shutdown()
	}

	manApp, err := NewManApp(assets.Manual, app)
	if err != nil {
		panic(err)
	}

	app.rootCmd.AddCommand(
		slices.Map(modules.Categories(), func(c modules.Category) *cobra.Command {
			return ExecuteModuleCommand(
				c,
				app,
				TasksArgsProviderFor(app, c),
				WithShort(fmt.Sprintf("Run a %s by its name", modules.CategoryName(c))),
			)
		})...,
	)

	//nolint:contextcheck // there's no context to pass here
	app.rootCmd.AddCommand(
		ModulesCommand(app, manApp, app),
		PluginsCommand(NewPluginApp(app, app.pluginMgr, app)),
		VaultCommand(NewVaultApp(app, app)),
		ServerCommand(NewServerApp(app)),
		EnvCommand(NewEnvApp(app)),
		RunCommand(app, app, app),
		VersionCommand(),
	)

	app.rootCmd.PersistentFlags().AddGoFlagSet(app.appCfg.Flags())
	app.rootCmd.PersistentFlags().AddGoFlagSet(app.loggingCfg.Flags())

	return app
}

var _ ModuleCommander = (*App)(nil)

type appServices struct {
	registry     *modules.TypeRegistry
	vault        *vault.Vault
	dockerClient *lazy.Lazy[*client.Client]
	ignorer      *ignore.Ignorer
	cache        state.Cache
	stateDB      *state.DB
	diagsWriter  hcl2.DiagnosticWriter
	repo         *lazy.Lazy[*modules.Repository]
}

type App struct {
	parsingState struct {
		parsingRemainder hcl2.Body
		currentEvalCtx   *hcl2.EvalContext
	}
	services    appServices
	vcs         *lazy.Lazy[any]
	repoDetails *lazy.Lazy[repo.Repo]
	buildrCfg   *lazy.Lazy[*config.Buildr]
	rootCmd     *cobra.Command
	recorder    *profiling.Recorder
	pluginMgr   *plugins.Manager
	loggingCfg  logging.Config
	appCfg      AppConfig
	toClose     []io.Closer
}

func (a *App) SetVault(v *vault.Vault) {
	a.services.vault = v
}

func (a *App) Vault() (*vault.Vault, error) {
	return a.services.vault, nil
}

func (a *App) PluginsRepo() (state.Plugins, error) {
	if a.services.stateDB == nil {
		return nil, ErrStateStoreNotInitialized
	}
	return a.services.stateDB.Plugins, nil
}

func (a *App) TypeRegistry() *modules.TypeRegistry {
	return a.services.registry
}

func (a *App) Run(ctx context.Context) error {
	return a.rootCmd.ExecuteContext(ctx)
}

func (a *App) RunWithArgs(ctx context.Context, args ...string) error {
	a.rootCmd.SetArgs(args)
	return a.rootCmd.ExecuteContext(ctx)
}

func (a *App) Init(ctx context.Context) (err error) {
	logger, closer, err := a.loggingCfg.Logger()
	if err != nil {
		return err
	}
	a.toClose = append(a.toClose, closer)
	slog.SetDefault(logger)

	if a.recorder, err = a.appCfg.Profiling.Setup(); err != nil {
		return err
	}

	if err := a.appCfg.InitPaths(cwd); err != nil {
		if !errors.Is(err, ErrFailedToFindDirectory) {
			return err
		}
	}

	if a.services.ignorer, err = a.appCfg.Ignorer(); err != nil {
		return err
	}

	if a.services.vault, err = a.appCfg.InitVault(); err != nil {
		return err
	}

	if a.services.stateDB, err = a.appCfg.InitState(ctx); err != nil {
		return err
	}

	a.services.cache = state.NewStateCache(a.appCfg.Cache.TTL, a.services.stateDB.State)

	a.pluginMgr.Init(a.services.stateDB.Plugins, a.appCfg.Cache.Directory)

	if err := a.pluginMgr.Register(ctx, a.services.registry); err != nil {
		return err
	}

	// init these in the background for better startup performance
	a.vcs = lazy.NewFuture[any](a.appCfg.ParseVCSInfo)
	a.repoDetails = lazy.NewFuture[repo.Repo](a.prepareRepoMetaInfo)

	a.buildrCfg = lazy.NewLazyCtx(ctx, a.prepareBuildrConfig)
	a.services.repo = lazy.NewLazyCtx(ctx, a.prepareModulesRepo)

	return nil
}

func (a *App) Shutdown() error {
	shutdownErrors := make([]error, 0)

	if err := a.persistVaultState(); err != nil {
		shutdownErrors = append(shutdownErrors, err)
	}

	a.toClose = append(a.toClose, a.recorder, a.services.dockerClient)
	for _, c := range a.toClose {
		if err := c.Close(); err != nil {
			shutdownErrors = append(shutdownErrors, c.Close())
		}
	}

	return errors.Join(shutdownErrors...)
}

func (a *App) AppConfig() AppConfig {
	return a.appCfg
}

func (a *App) BuildrConfig() (*config.Buildr, error) {
	return a.buildrCfg.Get()
}

func (a *App) Repository() (*modules.Repository, error) {
	return a.services.repo.Get()
}

func (a *App) BootstrapModule(_ context.Context, cat modules.Category, typeName, moduleName string) error {
	mod, err := a.services.registry.Create(cat, typeName)
	if err != nil {
		return err
	}

	mod.SetName(moduleName)

	writer := hcl.NewWriter[modules.ModuleWithMeta](os.Stdout)

	return writer.Write(mod)
}

func (a *App) RunModule(ctx context.Context, cat modules.Category, name string) error {
	binPathProvider, err := containers.NewDefaultServerBinPathProvider(nil, CurrentVersion)
	if err != nil {
		return fmt.Errorf("failed to create server bin path provider: %w", err)
	}

	orchestratorLazy := lazy.NewLazy(func() (*containers.Orchestrator, error) {
		cli, err := a.services.dockerClient.Get()
		if err != nil {
			return nil, err
		}
		return containers.NewOrchestrator(ctx, cli, a.services.ignorer, binPathProvider)
	})

	modulesRepo, err := a.services.repo.Get()
	if err != nil {
		return err
	}

	factory := execution.NewTaskFactory(
		execution.WithProvider(local.Provider(a.services.stateDB.State)),
		execution.WithProvider(container.Provider(orchestratorLazy, modulesRepo, a.services.stateDB.State)),
	)

	plan, err := execution.NewPlanFor(cat, name, modulesRepo, factory)
	if err != nil {
		return err
	}

	buildrCfg, err := a.buildrCfg.Get()
	if err != nil {
		return err
	}

	return plan.Execute(ctx, execution.Spec{
		ProjectRoot:       a.appCfg.ProjectRoot,
		CacheDirectory:    a.appCfg.Cache.Directory,
		BinariesDirectory: buildrCfg.BinariesDirectory,
		OutDirectory:      buildrCfg.OutDirectory,
		LogsDirectory:     buildrCfg.LogsDirectory,
		LogToStdErr:       buildrCfg.LogToStderr,
	})
}

func (a *App) RunTool(ctx context.Context, name string, args []string) error {
	buildrCfg, err := a.buildrCfg.Get()
	if err != nil {
		return err
	}
	var pathBuilder strings.Builder
	_, _ = pathBuilder.WriteString(buildrCfg.BinariesDirectory)

	currentPath, ok := os.LookupEnv("PATH")
	if ok {
		_, _ = pathBuilder.WriteRune(os.PathListSeparator)
		_, _ = pathBuilder.WriteString(currentPath)
	}

	if err := os.Setenv("PATH", pathBuilder.String()); err != nil {
		return err
	}

	modulesRepo, err := a.services.repo.Get()
	if err != nil {
		return err
	}

	mod := modulesRepo.Module(modules.CategoryTool, name)
	if mod == nil {
		return fmt.Errorf("no tool with name %s found", name)
	}

	if bn, ok := mod.Unwrap().(modules.BinaryNamer); ok {
		if binaryName, err := bn.BinaryName(ctx); err != nil {
			return err
		} else {
			name = binaryName
		}
	}

	cmd := exec.CommandContext(ctx, name, args...)

	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	cmd.Stdin = os.Stdin

	err = cmd.Run()
	if err == nil {
		return nil
	}

	var exitErr *exec.ExitError
	if errors.As(err, &exitErr) {
		slog.Default().Error(
			"Failed to run tool",
			slog.Bool("exited", exitErr.Exited()),
			slog.Int("exit_code", exitErr.ExitCode()),
			slog.Duration("duration", exitErr.UserTime()),
		)
		return errs.ErrAlreadyLogged
	}

	return err
}

func (a *App) prepareRepoMetaInfo() (repo.Repo, error) {
	if vi, err := a.vcs.Get(); err != nil {
		if errors.Is(err, ErrNoVCSRootDetected) {
			return repo.Repo{}, nil
		}
		return repo.Repo{}, err
	} else {
		return a.appCfg.CollectRepoDetails(vi)
	}
}

func (a *App) persistVaultState() error {
	instance := a.services.vault
	if !instance.IsInitialized() || instance.IsInMemory() {
		return nil
	}

	outFile, err := os.Create(a.appCfg.Vault.FilePath)
	if err != nil {
		return err
	}

	defer func() {
		err = errors.Join(err, outFile.Close())
	}()

	encoder := json.NewEncoder(outFile)

	return encoder.Encode(instance)
}

func (a *App) prepareBuildrConfig(ctx context.Context) (*config.Buildr, error) {
	buildrFS, err := a.appCfg.BuildRFS()
	if err != nil {
		return nil, err
	}

	parser := hcl.NewParser(slog.Default(), buildrFS)
	slog.Debug("Reading files", "buildr_directory", a.appCfg.BuildRDirectory)

	if err := parser.ReadFiles(ctx); err != nil {
		slog.Error("Failed to read files", err)
		return nil, err
	}

	buildrCfg := struct {
		Remainder      hcl2.Body `hcl:",remain"`
		*config.Buildr `hcl:"buildr,block"`
	}{
		Buildr: &config.Buildr{
			LogToStderr: a.appCfg.Execution.LogToStderr,
		},
	}

	a.parsingState.currentEvalCtx = hcl.BasicContext(a.services.vault)
	if err := parser.Parse(a.parsingState.currentEvalCtx, &buildrCfg); err != nil {
		return nil, err
	}

	a.services.diagsWriter = parser.DiagsWriter()

	buildrDir, err := a.appCfg.BuildRDirectory()
	if err != nil {
		return nil, err
	}

	if err := buildrCfg.SetupDirectories(buildrDir, a.appCfg.ProjectRoot, a.appCfg.Execution.CleanDirectories); err != nil {
		return nil, err
	}

	a.parsingState.parsingRemainder = buildrCfg.Remainder

	return buildrCfg.Buildr, nil
}

func (a *App) prepareModulesRepo(ctx context.Context) (*modules.Repository, error) {
	// call this early to ensure the parsing state is initialized
	buildrCfg, err := a.buildrCfg.Get()
	if err != nil {
		return nil, err
	}

	evalCtx := hcl.FullContext(ctx, a.parsingState.currentEvalCtx, a.services.vault, a.services.cache)
	var rawSpec hcl.ModulesSpec

	if diags := gohcl.DecodeBody(a.parsingState.parsingRemainder, evalCtx, &rawSpec); diags.HasErrors() {
		_ = a.services.diagsWriter.WriteDiagnostics(diags)
		return nil, errs.ErrAlreadyLogged
	}

	var repoOptions []hcl.RepositoryOption

	vcsInfo, err := a.vcs.Get()
	if err == nil {
		repoOptions = append(repoOptions, hcl.WithGlobalVariable("vcs", vcsInfo))
	}

	repoDetails, err := a.repoDetails.Get()
	if err != nil {
		return nil, err
	} else {
		repoOptions = append(repoOptions, hcl.WithGlobalVariable("repo", repoDetails))
	}

	return rawSpec.Repository(
		evalCtx,
		a.services.diagsWriter,
		a.services.registry,
		buildrCfg.OutDirectory,
		repoOptions...,
	)
}

func prepareDockerAPIClient(ctx context.Context) (*client.Client, error) {
	const dockerClientCloseTimeout = 200 * time.Millisecond

	cli, err := client.NewClientWithOpts(
		client.WithHostFromEnv(),
		client.WithVersionFromEnv(),
		client.WithTLSClientConfigFromEnv(),
	)
	if err != nil {
		return nil, err
	}

	dockerCtx, dockerCancel := context.WithTimeout(ctx, dockerClientCloseTimeout)
	defer dockerCancel()

	cli.NegotiateAPIVersion(dockerCtx)

	return cli, nil
}
