package cmd

import (
	"context"
	"fmt"
	"io"
	"io/fs"
	"log/slog"
	"text/template"

	"github.com/Masterminds/sprig/v3"

	"code.icb4dc0.de/buildr/buildr/internal/tmpl"
	"code.icb4dc0.de/buildr/buildr/modules"
)

var _ ManCommander = (*ManApp)(nil)

type ManAppServiceAccess interface {
	TypeRegistryAccessor
}

type Pager interface {
	io.Writer
	Display() error
}

func NewManApp(templatesFs fs.FS, svcAcc ManAppServiceAccess) (*ManApp, error) {
	parsedTemplates, err := template.New("man").
		Funcs(sprig.FuncMap()).
		Funcs(template.FuncMap{
			"to_hcl": tmpl.WriteToHcl,
		}).
		ParseFS(templatesFs, "manual/*.tmpl.md")
	if err != nil {
		return nil, err
	}

	return &ManApp{
		templates:     parsedTemplates,
		serviceAccess: svcAcc,
	}, nil
}

type ManApp struct {
	templates     *template.Template
	serviceAccess ManAppServiceAccess
}

func (m ManApp) DisplayModuleManual(ctx context.Context, pager Pager, cat modules.Category, name string) error {
	mod, err := m.serviceAccess.TypeRegistry().Create(cat, name)
	if err != nil {
		return err
	}

	var help modules.Help
	if h, ok := mod.Unwrap().(modules.Helper); !ok {
		slog.Info(
			"Module has no help",
			slog.String("category", modules.CategoryName(cat)),
			slog.String("name", name),
		)
		return nil
	} else if he, err := h.Help(ctx); err == nil {
		help = he
	} else {
		return fmt.Errorf("failed to get module help: %w", err)
	}

	if err := m.templates.ExecuteTemplate(pager, "single-module-man.tmpl.md", help); err != nil {
		return err
	}

	return pager.Display()
}

func (m ManApp) DisplayModulesManual(pager Pager) error {
	templateData := struct {
		Modules map[modules.Category][]string
	}{
		Modules: m.serviceAccess.TypeRegistry().Inventory(),
	}

	if err := m.templates.ExecuteTemplate(pager, "modules-man.tmpl.md", templateData); err != nil {
		return err
	}

	return pager.Display()
}
