package cmd

import (
	"context"
	"errors"
	"fmt"
	"io"
	"log/slog"
	"os"

	"code.icb4dc0.de/prskr/go-pwgen"
)

const defaultPassphraseFileMode os.FileMode = 0o600

var _ VaultCommander = (*VaultApp)(nil)

type VaultAppServiceAccess interface {
	VaultAccessor
	VaultModifier
}

func NewVaultApp(
	serviceAccess VaultAppServiceAccess,
	configAccessor AppConfigAccessor,
) *VaultApp {
	return &VaultApp{
		serviceAccess:  serviceAccess,
		configAccessor: configAccessor,
	}
}

type VaultApp struct {
	serviceAccess  VaultAppServiceAccess
	configAccessor AppConfigAccessor
}

func (v VaultApp) Init(_ context.Context, initCfg VaultInitConfig) error {
	vault, err := v.serviceAccess.Vault()
	if err != nil {
		return err
	}

	if vault != nil {
		slog.Default().Info("vault already initialized")
		return nil
	}

	cfg := v.configAccessor.AppConfig()

	if cfg.Vault.PassphraseFile == "" {
		return errors.New("vault.passphrase-file may not be empty")
	}

	key, err := pwgen.Generate(pwgen.WithLength(initCfg.PassphraseLength))
	if err != nil {
		return fmt.Errorf("error generating password: %w", err)
	}

	if err := os.WriteFile(cfg.Vault.PassphraseFile, []byte(key), defaultPassphraseFileMode); err != nil {
		return fmt.Errorf("failed to write passphrase to file: %w", err)
	}

	if vault, err := cfg.InitVault(); err != nil {
		return fmt.Errorf("failed to prepare vault: %w", err)
	} else {
		v.serviceAccess.SetVault(vault)
	}

	return nil
}

func (v VaultApp) List(_ context.Context, writer io.Writer) error {
	vaultInstance, err := v.serviceAccess.Vault()
	if err != nil {
		return err
	}

	if vaultInstance == nil {
		return ErrVaultNotInitiated
	}

	_, _ = fmt.Fprintln(writer)

	for _, k := range vaultInstance.Keys() {
		if _, err := fmt.Fprintf(writer, "\t%s\n", k); err != nil {
			return err
		}
	}

	return nil
}

func (v VaultApp) Get(_ context.Context, key string, writer io.Writer) error {
	vaultInstance, err := v.serviceAccess.Vault()
	if err != nil {
		return err
	}
	if vaultInstance == nil {
		return ErrVaultNotInitiated
	}

	val, err := vaultInstance.GetValue(key)
	if err != nil {
		return err
	}

	_, _ = fmt.Fprintln(writer)
	if _, err := fmt.Fprintln(writer, val); err != nil {
		return fmt.Errorf("failed to print value: %w", err)
	}

	return nil
}

func (v VaultApp) Set(_ context.Context, key string, value []byte) error {
	vaultInstance, err := v.serviceAccess.Vault()
	if err != nil {
		return err
	}
	if vaultInstance == nil {
		return ErrVaultNotInitiated
	}

	return vaultInstance.SetValue(key, value)
}

func (v VaultApp) Remove(_ context.Context, key string) error {
	vaultInstance, err := v.serviceAccess.Vault()
	if err != nil {
		return err
	}
	if vaultInstance == nil {
		return ErrVaultNotInitiated
	}

	vaultInstance.RemoveValue(key)

	return nil
}
