package cmd

import (
	"context"
	"errors"
	"flag"
	"fmt"
	"io/fs"
	"os"
	"path/filepath"
	"strconv"
	"time"

	"code.icb4dc0.de/buildr/buildr/internal/semver"
	"code.icb4dc0.de/buildr/buildr/modules/repo"

	"code.icb4dc0.de/buildr/buildr/modules/state"

	"code.icb4dc0.de/buildr/buildr/internal/config"
	"code.icb4dc0.de/buildr/buildr/internal/ignore"
	"code.icb4dc0.de/buildr/buildr/internal/profiling"
	"code.icb4dc0.de/buildr/buildr/internal/vault"
	"code.icb4dc0.de/buildr/buildr/modules/vcs"
)

const (
	defaultCacheTTL             = 15 * time.Minute
	buildrDirectoryName         = ".buildr"
	buildrProjectRootEnv        = "BUILDR_PROJECT_ROOT"
	defaultDirectoryPermissions = 0o755
)

var (
	ErrNoProjectDir          = errors.New("no project directory detected")
	ErrNoVCSRootDetected     = errors.New("no VCS root detected")
	ErrFailedToFindDirectory = errors.New("failed to find directory")
)

type AppConfig struct {
	Vault struct {
		FilePath       string
		Passphrase     string
		PassphraseFile string
	}
	ProjectRoot string
	VCSRoot     string
	VCSType     vcs.Type
	State       struct{ FilePath string }
	Profiling   profiling.Config
	Cache       struct {
		Directory string
		TTL       time.Duration
	}
	Execution struct {
		LogToStderr      bool
		CleanDirectories bool
	}
}

func (c *AppConfig) CollectRepoDetails(vcsInfo any) (repoDetails repo.Repo, err error) {
	if c.VCSRoot == "" || vcsInfo == nil {
		return repoDetails, nil
	}

	repoDetails.Root = c.VCSRoot

	if git, ok := vcsInfo.(*vcs.Git); ok && git.Tag != "" && semver.IsValid(git.Tag) {
		if repoDetails.Version, err = semver.ParseVersion(git.Tag); err != nil {
			return repoDetails, err
		} else {
			return repoDetails, nil
		}
	}

	return repoDetails, nil
}

func (c *AppConfig) ParseVCSInfo() (vcsDetails any, err error) {
	if c.VCSRoot == "" {
		return nil, ErrNoVCSRootDetected
	}

	switch c.VCSType {
	case vcs.TypeGit:
		gitInfo, err := vcs.ParseGitInfo(c.VCSRoot)
		if err != nil {
			return nil, err
		}

		return gitInfo, nil
	default:
		return nil, fmt.Errorf("unsupported VCS type: %s", c.VCSType)
	}
}

func (c *AppConfig) BuildRFS() (fs.FS, error) {
	if dir, err := c.BuildRDirectory(); err != nil {
		return nil, err
	} else {
		return os.DirFS(dir), nil
	}
}

func (c *AppConfig) BuildRDirectory() (string, error) {
	if c.ProjectRoot == "" {
		return "", ErrNoProjectDir
	}
	return filepath.Join(c.ProjectRoot, buildrDirectoryName), nil
}

func (c *AppConfig) Ignorer() (*ignore.Ignorer, error) {
	if c.ProjectRoot == "" {
		return ignore.NewIgnorer(cwd)
	}
	return ignore.NewIgnorer(c.ProjectRoot, c.Vault.FilePath)
}

func (c *AppConfig) InitPaths(from string) (err error) {
	if ucd, err := os.UserCacheDir(); err != nil {
		return err
	} else {
		c.Cache.Directory = filepath.Join(ucd, "buildr")
		if err := os.MkdirAll(c.Cache.Directory, defaultDirectoryPermissions); err != nil {
			return err
		}
	}

	var projectRoot string
	if root, ok := os.LookupEnv(buildrProjectRootEnv); ok {
		projectRoot = root
	} else if root, err = c.findDirectory(from, buildrDirectoryName); err != nil {
		return err
	} else {
		projectRoot = root
	}

	c.ProjectRoot = projectRoot

	if d, err := c.findDirectory(c.ProjectRoot, c.VCSType.Directory()); err != nil && !errors.Is(err, ErrFailedToFindDirectory) {
		return err
	} else {
		c.VCSRoot = d
	}

	if stateFilePath := c.State.FilePath; stateFilePath != "" && !filepath.IsAbs(stateFilePath) {
		c.State.FilePath = filepath.Join(c.ProjectRoot, stateFilePath)
	}

	return nil
}

func (c *AppConfig) InitState(ctx context.Context) (*state.DB, error) {
	if stateFilePath := c.State.FilePath; c.ProjectRoot != "" && stateFilePath != "" {
		return state.NewDB(ctx, c.State.FilePath)
	}

	return state.NewInMemDB(ctx)
}

func (c *AppConfig) InitVault() (*vault.Vault, error) {
	var vaultOpts []vault.InitOption

	if pw := c.Vault.Passphrase; pw != "" {
		vaultOpts = append(vaultOpts, vault.WithPassphrase(pw))
	} else if path := c.Vault.PassphraseFile; path != "" {
		vaultOpts = append(vaultOpts, vault.WithPassphraseFile(path))
	} else {
		return vault.NewInMemVault()
	}

	inst, err := vault.NewVault(vaultOpts...)
	if err != nil {
		return nil, err
	}

	if filePath := c.Vault.FilePath; c.ProjectRoot != "" && filePath != "" {
		if !filepath.IsAbs(c.Vault.FilePath) {
			filePath = filepath.Join(c.ProjectRoot, filePath)
		}

		if err := inst.LoadFromFile(filePath); err != nil {
			return nil, err
		}
	}

	return inst, nil
}

func (c *AppConfig) Flags() *flag.FlagSet {
	flags := flag.NewFlagSet("buildr", flag.ExitOnError)

	c.VCSType = config.EnvOr("BUILDR_VCS_TYPE", vcs.ParseType, vcs.TypeGit)

	flags.StringVar(
		&c.Vault.FilePath,
		"vault.file",
		config.StringEnvOr("BUILDR_VAULT_FILE", filepath.Join(".buildr", "vault.json")),
		"Relative file path to vault file",
	)

	flags.StringVar(
		&c.Vault.Passphrase,
		"vault.passphrase",
		config.StringEnvOr("BUILDR_VAULT_PASSPHRASE", ""),
		"Password for vault - has precedence over vault-passphrase-file",
	)

	flags.StringVar(
		&c.Vault.PassphraseFile,
		"vault.passphrase-file",
		config.StringEnvOr("BUILDR_VAULT_PASSPHRASE_FILE", filepath.Join(".buildr", ".vaultpw")),
		"File containing the vault passphrase",
	)

	flags.BoolVar(
		&c.Execution.LogToStderr,
		"execution.log-to-stderr",
		config.EnvOr("BUILDR_EXECUTION_LOG_TO_STDERR", strconv.ParseBool, false),
		"If output of commands should be piped to stdout",
	)

	flags.BoolVar(
		&c.Execution.CleanDirectories,
		"execution.clean-directories",
		config.EnvOr("BUILDR_EXECUTION_CLEAN_DIRECTORIES", strconv.ParseBool, false),
		"If ephemeral directories (currently out and logs) should be cleaned",
	)

	flags.StringVar(
		&c.State.FilePath,
		"state.file-path",
		config.StringEnvOr("BUILDR_STATE_FILE_PATH", filepath.Join(".buildr", "state.sqlite")),
		"Relative file path to state file",
	)

	flags.DurationVar(
		&c.Cache.TTL,
		"cache.ttl",
		config.EnvOr("BUILDR_CACHE_TTL", time.ParseDuration, defaultCacheTTL),
		"TTL for cache entries",
	)

	flags.Var(
		&c.VCSType,
		"vcs.type",
		"The type of VCS to use",
	)

	c.Profiling.AddFlags(flags)

	return flags
}

func (c *AppConfig) findDirectory(from, dirToSearch string) (string, error) {
	if from == "." || from == "" || from == "/" {
		return "", fmt.Errorf("%w: %s", ErrFailedToFindDirectory, dirToSearch)
	}

	info, err := os.Stat(filepath.Join(from, dirToSearch))
	if err != nil {
		return c.findDirectory(filepath.Dir(from), dirToSearch)
	}

	if info.IsDir() {
		return from, nil
	}

	return c.findDirectory(filepath.Dir(from), dirToSearch)
}
