package containers

import (
	"context"
	"errors"
	"fmt"
	"net/http"
	"os"
	"path/filepath"
	"runtime"

	"github.com/klauspost/pgzip"

	"code.icb4dc0.de/buildr/buildr/internal/ioutils"
)

const executableMode os.FileMode = 0o755

type ServerPlatform struct {
	OS   string
	Arch string
}

type HTTPClient interface {
	Do(req *http.Request) (*http.Response, error)
}

type ServerBinPathProvider interface {
	ServerBinPath(ctx context.Context, platform ServerPlatform) (string, error)
}

func NewDefaultServerBinPathProvider(client HTTPClient, currentVersion string) (ServerBinPathProvider, error) {
	if client == nil {
		client = http.DefaultClient
	}

	cacheDir, err := os.UserCacheDir()
	if err != nil {
		return nil, fmt.Errorf("failed to get user cache dir: %w", err)
	}

	return &DefaultServerBinPathProvider{
		client:         client,
		cacheDir:       cacheDir,
		currentVersion: currentVersion,
		currentPlatform: ServerPlatform{
			OS:   runtime.GOOS,
			Arch: runtime.GOARCH,
		},
	}, nil
}

type DefaultServerBinPathProvider struct {
	client          HTTPClient
	currentVersion  string
	cacheDir        string
	currentPlatform ServerPlatform
}

func (p DefaultServerBinPathProvider) ServerBinPath(ctx context.Context, platform ServerPlatform) (binaryPath string, err error) {
	if p.currentPlatform == platform {
		return os.Executable()
	}

	binaryPath = filepath.Join(p.cacheDir, "buildr", fmt.Sprintf("buildr_%s_%s_%s", p.currentVersion, platform.OS, platform.Arch))

	if _, err = os.Stat(binaryPath); err == nil {
		return binaryPath, nil
	}

	req, err := http.NewRequestWithContext(
		ctx,
		http.MethodGet,
		fmt.Sprintf("https://code.icb4dc0.de/buildr/buildr/releases/download/%s/buildr_%s_%s.gz", p.currentVersion, platform.OS, platform.Arch),
		nil,
	)
	if err != nil {
		return "", fmt.Errorf("failed to prepare binary request: %w", err)
	}

	resp, err := p.client.Do(req)
	if err != nil {
		return "", fmt.Errorf("failed to download binary: %w", err)
	}

	defer func() {
		err = errors.Join(err, resp.Body.Close())
	}()

	if resp.StatusCode != http.StatusOK {
		return "", fmt.Errorf("failed to download binary: %s", resp.Status)
	}

	outFile, err := os.OpenFile(binaryPath, os.O_CREATE|os.O_EXCL|os.O_WRONLY, executableMode)
	if err != nil {
		return "", fmt.Errorf("failed to create binary file: %w", err)
	}

	defer func() {
		err = errors.Join(err, outFile.Close())
	}()

	gzipReader, err := pgzip.NewReader(resp.Body)
	if err != nil {
		return "", fmt.Errorf("failed to create gzip reader: %w", err)
	}

	_, err = ioutils.CopyWithPooledBuffer(outFile, gzipReader)

	return binaryPath, err
}
