package local

import (
	"context"
	"errors"
	"fmt"
	"log/slog"
	"os"
	"path/filepath"
	"sync"

	"code.icb4dc0.de/buildr/buildr/modules/state"

	"golang.org/x/sync/errgroup"

	"code.icb4dc0.de/buildr/buildr/internal/execution"
	"code.icb4dc0.de/buildr/buildr/internal/storage"
	"code.icb4dc0.de/buildr/buildr/modules"
)

const defaultDirectoryMode os.FileMode = 0o755

var _ execution.Task = (*localTask)(nil)

type localTask struct {
	module     modules.ModuleWithMeta
	stateStore state.Store
	execution.TaskDependencies
	once sync.Once
}

func (t *localTask) Execute(ctx context.Context, spec execution.Spec) (err error) {
	t.once.Do(func() {
		err = t.doExecute(ctx, spec)
	})

	return err
}

func (t *localTask) doExecute(ctx context.Context, spec execution.Spec) error {
	if err := t.TaskDependencies.Execute(ctx, spec); err != nil {
		return err
	}

	defer func() {
		if rec := recover(); rec != nil {
			slog.Error("Recovered panic during execution", slog.Any("recovered_value", rec))
		}
	}()

	if outDir := t.module.OutDir(); outDir != "" {
		if err := os.MkdirAll(t.module.OutDir(), defaultDirectoryMode); err != nil {
			return err
		}
	}

	if mappings := t.module.InputMappings(); len(mappings) > 0 || t.module.Category() == modules.CategoryTool {
		return t.executeIsolated(ctx, spec)
	}

	execCtx, err := t.executionContextFor(
		ctx,
		t.module,
		spec,
		spec.ProjectRoot,
		t.module.OutDir(),
	)
	if err != nil {
		return fmt.Errorf("failed to prepare execution context: %w", err)
	}

	defer func() {
		err = errors.Join(err, execCtx.Close())
	}()

	if t.module.ShouldSkip() {
		execCtx.Logger().Info("Skipping execution as requested")
		return nil
	}

	if err := t.module.Execute(execCtx); err != nil {
		return fmt.Errorf("failed to execute module %s/%s: %w", modules.CategoryName(t.module.Category()), t.module.Name(), err)
	}

	return nil
}

func (t *localTask) executeIsolated(ctx context.Context, spec execution.Spec) error {
	workDir, err := os.MkdirTemp(os.TempDir(), "buildr-work-*")
	if err != nil {
		return err
	}

	defer func() {
		_ = os.RemoveAll(workDir)
	}()

	outDir := t.module.OutDir()
	if t.module.Category() == modules.CategoryTool {
		outDir = spec.BinariesDirectory
	}

	mappings := t.module.InputMappings()
	layers := make(storage.Layers, 0, len(mappings))

	for k, v := range mappings {
		layers = append(layers, storage.Layer{
			Source: k,
			Target: v,
		})
	}

	if !filepath.IsAbs(outDir) {
		outDir = filepath.Join(spec.ProjectRoot, outDir)
	}

	ufs := storage.NewUnionFS(outDir)
	srv, err := storage.MountUnionFS(ufs, workDir, layers)
	if err != nil {
		return err
	}

	grp, grpcCtx := errgroup.WithContext(ctx)

	grp.Go(func() error {
		srv.Wait()
		return nil
	})

	if err := srv.WaitMount(); err != nil {
		return err
	}

	defer func() {
		if err := srv.Unmount(); err != nil {
			slog.Error("failed to unmount unionfs", slog.Any("error", err))
		}
	}()

	grp.Go(func() (err error) {
		defer func() {
			err = errors.Join(err, srv.Unmount())
		}()

		execCtx, err := t.executionContextFor(
			grpcCtx,
			t.module,
			spec,
			workDir,
			outDir,
		)
		if err != nil {
			return err
		}

		defer func() {
			err = errors.Join(err, execCtx.Close())
		}()

		if err = t.module.Execute(execCtx); err != nil {
			return fmt.Errorf("failed to execute module %s/%s: %w", modules.CategoryName(t.module.Category()), t.module.Name(), err)
		}

		return nil
	})

	return grp.Wait()
}

func (t *localTask) executionContextFor(
	ctx context.Context,
	m modules.ModuleWithMeta,
	spec execution.Spec,
	workingDir, outDir string,
) (DefaultExecutionContext, error) {
	return NewDefaultExecutionContext(
		ctx,
		t.stateStore,
		m,
		workingDir,
		outDir,
		spec.BinariesDirectory,
		spec.LogsDirectory,
		spec.LogToStdErr,
		WithLoggerFactory(func() *slog.Logger {
			return slog.Default().With(
				slog.String("module_name", m.Name()),
				slog.String("module_type", modules.CategoryName(m.Category())),
			)
		}),
	)
}
