package hcl

import (
	"context"
	"errors"
	"fmt"
	"io/fs"
	"log/slog"
	"os"
	"path/filepath"
	"sync"

	"github.com/hashicorp/hcl/v2"
	"github.com/hashicorp/hcl/v2/gohcl"
	"github.com/hashicorp/hcl/v2/hclparse"

	"code.icb4dc0.de/buildr/buildr/internal/errs"
	"code.icb4dc0.de/buildr/buildr/internal/maps"
)

func NewParser(logger *slog.Logger, inputFS fs.FS) *Parser {
	return &Parser{
		logger: logger,
		fs:     inputFS,
		parser: hclparse.NewParser(),
	}
}

type Parser struct {
	logger      *slog.Logger
	fs          fs.FS
	diagsWriter hcl.DiagnosticWriter
	parser      *hclparse.Parser
	Body        hcl.Body
}

func (p *Parser) DiagsWriter() hcl.DiagnosticWriter {
	return p.diagsWriter
}

func (p *Parser) Parse(ctx *hcl.EvalContext, val any) error {
	p.diagsWriter = hcl.NewDiagnosticTextWriter(os.Stderr, p.parser.Files(), 0, true)
	diags := gohcl.DecodeBody(p.Body, ctx, val)
	if diags.HasErrors() {
		_ = p.diagsWriter.WriteDiagnostics(diags)
		return errs.ErrAlreadyLogged
	}

	return nil
}

func (p *Parser) ReadFiles(ctx context.Context) (err error) {
	p.parser = hclparse.NewParser()

	type fileToBeParsed struct {
		path string
		data []byte
	}

	var (
		wg               sync.WaitGroup
		toBeParsed       = make(chan fileToBeParsed)
		processingErrors = make(chan error)
	)

	processingCtx, cancel := context.WithCancel(ctx)
	defer cancel()

	go func() {
		for {
			select {
			case <-processingCtx.Done():
				return
			case f := <-toBeParsed:
				if _, err := p.parser.ParseHCL(f.data, f.path); err.HasErrors() {
					_ = hcl.NewDiagnosticTextWriter(os.Stderr, p.parser.Files(), 0, true).WriteDiagnostics(err)
					processingErrors <- fmt.Errorf("parsing failed for file %s", f.path)
				}
				wg.Done()
			}
		}
	}()

	err = fs.WalkDir(p.fs, ".", func(path string, d fs.DirEntry, err error) error {
		if d.IsDir() {
			return nil
		}

		if filepath.Ext(path) != ".hcl" {
			return nil
		}

		wg.Add(1)
		go func() {
			if content, err := fs.ReadFile(p.fs, path); err != nil {
				processingErrors <- fmt.Errorf("failed to read file %s: %w", path, err)
			} else {
				ftbp := fileToBeParsed{
					path: path,
					data: content,
				}

				toBeParsed <- ftbp
			}
		}()

		return nil
	})

	if err != nil {
		return err
	}

	go func() {
		wg.Wait()
		cancel()
	}()

	go func() {
		for {
			select {
			case <-processingCtx.Done():
				return
			case e := <-processingErrors:
				err = errors.Join(err, e)
			}
		}
	}()

	<-processingCtx.Done()

	if err != nil {
		return err
	}

	p.Body = hcl.MergeFiles(maps.Values(p.parser.Files()))

	return nil
}
