package lazy

import (
	"context"
	"errors"
	"fmt"
	"io"
	"reflect"
	"sync"
)

var ErrEmptyLazy = errors.New("lazy is empty")

func NewFuture[T any](provider func() (T, error)) *Lazy[T] {
	l := NewLazy(provider)

	go func() {
		_, _ = l.Get()
	}()

	return l
}

func NewLazyCtx[T any](ctx context.Context, provider func(ctx context.Context) (T, error)) *Lazy[T] {
	return &Lazy[T]{
		baseContext: ctx,
		provider:    provider,
	}
}

func NewLazy[T any](provider func() (T, error)) *Lazy[T] {
	return &Lazy[T]{
		baseContext: context.Background(),
		provider: func(ctx context.Context) (T, error) {
			return provider()
		},
	}
}

type Lazy[T any] struct {
	out         T
	err         error
	baseContext context.Context
	provider    func(ctx context.Context) (T, error)
	once        sync.Once
}

func (l *Lazy[T]) Get() (v T, e error) {
	if l == nil {
		var t T
		return v, fmt.Errorf("%w type: %T", ErrEmptyLazy, t)
	}
	l.once.Do(func() {
		ctx, cancel := context.WithCancel(l.baseContext)
		defer cancel()
		l.out, l.err = l.provider(ctx)
	})

	return l.out, l.err
}

func (l *Lazy[T]) Set(t T) {
	l.provider = func(ctx context.Context) (T, error) {
		return t, nil
	}
	l.out = t
}

func (l *Lazy[T]) MustGet() (v T) {
	var err error
	v, err = l.Get()
	if err != nil {
		if errors.Is(err, ErrEmptyLazy) {
			return v
		}

		panic(err)
	}

	return v
}

func (l *Lazy[T]) Close() error {
	v := any(l.out)
	if v == nil {
		return nil
	}

	if c, ok := v.(io.Closer); ok && !reflect.ValueOf(c).IsNil() {
		return c.Close()
	}

	return nil
}
