package logging

import (
	"flag"
	"io"
	"log/slog"
	"os"
	"strings"

	"code.icb4dc0.de/buildr/buildr/internal/ioutils"
)

var _ flag.Value = (*LogOutput)(nil)

type LogOutput string

func (o *LogOutput) Set(s string) error {
	*o = LogOutput(s)
	return nil
}

func (o *LogOutput) String() string {
	return string(*o)
}

func (o *LogOutput) Open() (io.WriteCloser, error) {
	switch strings.ToLower(o.String()) {
	case "", "stderr":
		return ioutils.NoOpWriteCloser(os.Stderr), nil
	case "stdout":
		return ioutils.NoOpWriteCloser(os.Stdout), nil
	default:
		return os.Create(string(*o))
	}
}

const (
	LogOutputStderr LogOutput = "stderr"
	LogOutputStdout LogOutput = "stdout"
)

func NewConfig() Config {
	return Config{
		LogLevel:  new(slog.LevelVar),
		LogOutput: LogOutputStderr,
	}
}

type Config struct {
	LogLevel  *slog.LevelVar
	LogOutput LogOutput
	AddSource bool
}

func (c *Config) Flags() *flag.FlagSet {
	flagSet := flag.NewFlagSet("logging", flag.PanicOnError)
	flagSet.TextVar(
		c.LogLevel,
		"log.level",
		c.LogLevel,
		"set log level",
	)
	flagSet.BoolVar(
		&c.AddSource,
		"log.add-source",
		false,
		"Enable to get detailed information where the log was produced",
	)

	flagSet.Var(
		&c.LogOutput,
		"log.output",
		"Configure where the log output should be written to [stderr, stdout, any file path]",
	)

	return flagSet
}

func (c *Config) Logger() (*slog.Logger, io.Closer, error) {
	slogOptions := slog.HandlerOptions{
		AddSource: c.AddSource,
		Level:     c.LogLevel.Level(),
	}

	out, err := c.LogOutput.Open()
	if err != nil {
		return nil, nil, err
	}

	return slog.New(slog.NewTextHandler(out, &slogOptions)), out, nil
}
