package rpc

import (
	"context"
	"log/slog"
	"net"
	"net/http"
	"sync"
	"time"

	"buf.build/gen/go/grpc/grpc/connectrpc/go/grpc/health/v1/healthv1connect"
	"connectrpc.com/grpcreflect"
	"golang.org/x/net/http2"
	"golang.org/x/net/http2/h2c"

	"code.icb4dc0.de/buildr/api/generated/remote/v1/rpcv1connect"

	"code.icb4dc0.de/buildr/buildr/modules"

	v1 "code.icb4dc0.de/buildr/buildr/internal/rpc/v1"
)

const (
	serverShutdownTimeout   = 5 * time.Second
	defaultReadWriteTimeout = 200 * time.Millisecond
	defaultIdleTimeout      = 30 * time.Second
)

type TypeRegistryAccessor interface {
	TypeRegistry() *modules.TypeRegistry
}

func NewServer(
	logger *slog.Logger,
	typeRegistryAccess TypeRegistryAccessor,
) *BuildrAPI {
	if logger == nil {
		logger = slog.Default()
	}

	return &BuildrAPI{
		logger:             logger,
		typeRegistryAccess: typeRegistryAccess,
	}
}

type BuildrAPI struct {
	typeRegistryAccess TypeRegistryAccessor
	logger             *slog.Logger
	server             *http.Server
	serverRunning      chan struct{}
	lock               sync.Mutex
}

func (a *BuildrAPI) Start(ctx context.Context, address string) (err error) {
	a.lock.Lock()
	a.serverRunning = make(chan struct{})
	a.lock.Unlock()

	var lis net.Listener
	if lis, err = net.Listen("tcp", address); err != nil {
		return err
	}

	mux := http.NewServeMux()

	a.server = &http.Server{
		Handler:           h2c.NewHandler(mux, &http2.Server{}),
		ReadHeaderTimeout: defaultReadWriteTimeout,
		IdleTimeout:       defaultIdleTimeout,
		BaseContext: func(net.Listener) context.Context {
			return ctx
		},
	}

	reflector := grpcreflect.NewStaticReflector(
		"grpc.health.v1.Health",
		"buildr.rpc.v1.ExecutorService",
		// "buildr.rpc.v1.buildr.rpc.v1.PluginManagerService",
	)

	mux.Handle(grpcreflect.NewHandlerV1Alpha(reflector))
	mux.Handle(grpcreflect.NewHandlerV1(reflector))
	mux.Handle(healthv1connect.NewHealthHandler(v1.NewHealthServer(a.logger)))
	mux.Handle(rpcv1connect.NewExecutorServiceHandler(v1.NewExecutorServiceServer(a.typeRegistryAccess.TypeRegistry())))

	defer func() {
		a.lock.Lock()
		close(a.serverRunning)
		a.lock.Unlock()
	}()

	return a.server.Serve(lis)
}

func (a *BuildrAPI) Close() {
	if !a.isRunning() {
		a.logger.Info("Skipping API server shutdown because server is not running")
		return
	}
	gracefulStopChan := make(chan struct{})
	go func() {
		ctx, cancel := context.WithTimeout(context.Background(), serverShutdownTimeout)
		defer cancel()
		_ = a.server.Shutdown(ctx)
		close(gracefulStopChan)
	}()

	select {
	case <-gracefulStopChan:
	case <-time.After(serverShutdownTimeout):
		_ = a.server.Close()
	}
}

func (a *BuildrAPI) isRunning() bool {
	a.lock.Lock()
	defer a.lock.Unlock()

	select {
	case _, more := <-a.serverRunning:
		return more
	default:
		return true
	}
}
