package v1

import (
	"context"
	"errors"
	"fmt"
	"log/slog"
	"os"
	"time"

	remotev1 "code.icb4dc0.de/buildr/api/generated/remote/v1"
	"code.icb4dc0.de/buildr/common/protocol"

	"code.icb4dc0.de/buildr/buildr/internal/archive"
	"code.icb4dc0.de/buildr/buildr/modules"

	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/status"
)

const (
	thresholdToWaitBeforeClosingWatcher             = 250 * time.Millisecond
	defaultDirectoryMode                os.FileMode = 0o755
)

type RemoteTaskExecutor struct {
	logger        *slog.Logger
	registry      *modules.TypeRegistry
	requestClient *RequestResponseClient
	sender        StreamSender[*remotev1.ExecutionServerMessage]
}

func (e *RemoteTaskExecutor) Execute(
	ctx context.Context,
	startTask *remotev1.StartTaskRequest,
) (resp chan *remotev1.TaskResult, errs chan error) {
	resp = make(chan *remotev1.TaskResult)
	errs = make(chan error)

	ctx, cancel := context.WithCancelCause(ctx)

	if err := os.MkdirAll(startTask.Buildr.Repo.Root, defaultDirectoryMode); err != nil {
		errs <- fmt.Errorf("unable to create working directory %q: %w", startTask.Buildr.Repo.Root, err)
		return resp, errs
	}

	go func() {
		defer cancel(errors.New("stream closed"))

		watcher, err := newFSWatcher(e.logger, startTask.Buildr.Repo.Root)
		if err != nil {
			e.logger.Error("Failed to start FS watcher", slog.String("err", err.Error()))
			errs <- status.Error(codes.Internal, err.Error())
			return
		}

		moduleRef := startTask.GetReference().GetModule()
		mod, err := e.registry.Create(moduleRef.GetModuleCategory(), moduleRef.GetModuleType())

		if err != nil {
			e.logger.Error("Failed to create module", slog.String("err", err.Error()))
			errs <- status.Error(codes.NotFound, err.Error())
			return
		} else if err := protocol.Unmarshal(startTask.Spec, mod); err != nil {
			e.logger.Error("Failed to unmarshal module spec", slog.String("err", err.Error()))
			errs <- status.Error(codes.InvalidArgument, err.Error())
			return
		} else {
			e.logger.Info("Executing module")
			go watcher.Watch(ctx)

			remoteState := newRemoteStateClient(e.logger, mod.Category(), startTask.Reference.GetName(), e.requestClient)

			executionContext := NewContainerExecutionContext(
				ctx,
				e.logger,
				startTask.Reference.Name,
				startTask.Buildr.Repo.Root,
				e.sender,
				remoteState,
			)

			err = mod.Execute(executionContext)
			result := new(remotev1.TaskResult)

			if err != nil {
				result.Error = fmt.Sprintf("failed to execute module %s/%s: %s", modules.CategoryName(mod.Category()), mod.Name(), err.Error())
			}

			e.logger.Debug("Waiting for file change events to propagate before canceling watcher")
			time.Sleep(thresholdToWaitBeforeClosingWatcher)

			cancel(ErrExecutionCompleted)

			if err := e.addModifiedFilesToResult(result, watcher); err != nil {
				errs <- err
				return
			}

			resp <- result
		}
	}()

	return resp, errs
}

func (e *RemoteTaskExecutor) addModifiedFilesToResult(result *remotev1.TaskResult, watcher *fsWatcher) error {
	tmpFile, err := os.CreateTemp(os.TempDir(), "buildr-modified-files-*.tar.s2")
	if err != nil {
		e.logger.Error("Failed to create temporary file", slog.String("err", err.Error()))
		return status.Error(codes.Internal, err.Error())
	}

	if err := watcher.WriteTo(tmpFile); err != nil {
		if errors.Is(err, archive.ErrEmptyArchive) {
			return nil
		}
		e.logger.Error("Failed to compress modified files", slog.String("err", err.Error()))
		return status.Error(codes.Internal, err.Error())
	}

	if err := tmpFile.Close(); err != nil {
		e.logger.Error("Failed to close temporary file", slog.String("err", err.Error()))
		return status.Error(codes.Internal, err.Error())
	}

	result.ModifiedFilesArchivePath = tmpFile.Name()
	return nil
}
