package v1

import (
	"context"
	"fmt"
	"io"
	"io/fs"
	"log/slog"
	"os"
	"path/filepath"

	"github.com/fsnotify/fsnotify"

	"code.icb4dc0.de/buildr/buildr/internal/archive"
	"code.icb4dc0.de/buildr/buildr/internal/ignore"
)

func newFSWatcher(logger *slog.Logger, workingDir string) (*fsWatcher, error) {
	watcher, err := fsnotify.NewWatcher()
	if err != nil {
		return nil, fmt.Errorf("failed to create fsnotify watcher: %w", err)
	}

	ignorer, err := ignore.NewIgnorer(workingDir)
	if err != nil {
		return nil, fmt.Errorf("failed to create ignorer for FSWatcher: %w", err)
	}

	w := &fsWatcher{
		logger:     logger,
		watcher:    watcher,
		workingDir: workingDir,
		ignorer:    ignorer,
		archive: archive.Tar{
			Ignorer: ignorer,
		},
	}

	if err := w.addDir(workingDir); err != nil {
		return nil, fmt.Errorf("failed to add working directory %q to FSWatcher: %w", workingDir, err)
	}

	return w, nil
}

type fsWatcher struct {
	logger     *slog.Logger
	ignorer    *ignore.Ignorer
	watcher    *fsnotify.Watcher
	archive    archive.Tar
	workingDir string
}

func (w *fsWatcher) WriteTo(writer io.Writer) error {
	return archive.WriteTarS2Encoded(w.archive, writer)
}

func (w *fsWatcher) Watch(ctx context.Context) {
	defer func() {
		if err := w.watcher.Close(); err != nil {
			w.logger.Error("Failed to close watcher", slog.String("err", err.Error()))
		}
	}()

	for {
		select {
		case <-ctx.Done():
			w.logger.Debug("Watcher is closing")
			return
		case err, ok := <-w.watcher.Errors:
			if !ok {
				return
			}
			w.logger.Error("Error occurred watching file changes", slog.String("err", err.Error()))
		case event, ok := <-w.watcher.Events:
			if !ok {
				return
			}

			switch event.Op {
			case fsnotify.Create, fsnotify.Write, fsnotify.Chmod:
				w.onCreate(event)
			case fsnotify.Rename:
				w.logger.Warn("Received rename event - ignored for now")
			case fsnotify.Remove:
				w.onDelete(event)
			}
		}
	}
}

func (w *fsWatcher) onCreate(event fsnotify.Event) {
	if info, err := os.Stat(event.Name); err != nil {
		w.logger.Error("Failed to stat file", slog.String("file", event.Name), slog.String("err", err.Error()))
		return
	} else if info.IsDir() {
		if err := w.addDir(event.Name); err != nil {
			w.logger.Error("Failed to add newly created directory to watcher", slog.String("file", event.Name), slog.String("err", err.Error()))
		}
	}

	relativeToWorkingDir, err := filepath.Rel(w.workingDir, event.Name)
	if err != nil {
		w.logger.Error(
			"Failed to make path relative to working directory",
			slog.String("path", event.Name),
			slog.String("err", err.Error()),
		)
		return
	}

	w.logger.Debug(
		"Processing file change",
		slog.String("orig_path", event.Name),
		slog.String("relative_path", relativeToWorkingDir),
	)

	if err := w.archive.Add(event.Name, relativeToWorkingDir); err != nil {
		w.logger.Error(
			"Failed to add file to archive",
			slog.String("path", relativeToWorkingDir),
			slog.String("err", err.Error()),
		)
		return
	}
}

func (w *fsWatcher) onDelete(event fsnotify.Event) {
	relativeToWorkingDir, err := filepath.Rel(w.workingDir, event.Name)
	if err != nil {
		w.logger.Error(
			"Failed to make path relative to working directory",
			slog.String("path", event.Name),
			slog.String("err", err.Error()),
		)
		return
	}

	if ok := w.archive.Remove(relativeToWorkingDir); !ok {
		w.logger.Warn(
			"Removed file is not part of archive - ignored for now",
			slog.String("path", relativeToWorkingDir),
		)
		return
	}
}

func (w *fsWatcher) addDir(dirToAdd string) (err error) {
	return w.ignorer.WalkDir(dirToAdd, func(path string, d fs.DirEntry, err error) error {
		if err != nil {
			return err
		}

		if !filepath.IsAbs(path) {
			path = filepath.Join(w.workingDir, path)
		}

		if d.IsDir() {
			w.logger.Debug("Subscribing to changes in directory", slog.String("directory", path))
			if err := w.watcher.Add(path); err != nil {
				return fmt.Errorf("failed to subscribe to changes in directory %s: %w", path, err)
			}
		}

		return nil
	})
}
