//go:build linux || darwin

package storage_test

import (
	"os"
	"path/filepath"
	"testing"

	"github.com/stretchr/testify/assert"

	"code.icb4dc0.de/buildr/buildr/internal/storage"
)

func TestUnionFS_CreateFile(t *testing.T) {
	t.Parallel()
	mountDir, outDir := setupFS(t)

	testDir := filepath.Join(mountDir, "work", "test")
	if err := os.MkdirAll(testDir, 0o755); err != nil {
		t.Fatalf("failed to create dir")
	}

	f, err := os.CreateTemp(testDir, "buildr-*")
	if err != nil {
		t.Fatalf("Failed to create file: %v", err)
	}
	_, testFileName := filepath.Split(f.Name())
	_, _ = f.WriteString("hello world")
	_ = f.Close()

	files, err := os.ReadDir(testDir)
	if err != nil {
		t.Fatalf("Failed to list files in dir %s: %v", testDir, err)
	}

	var foundFile bool
	for i := range files {
		foundFile = foundFile || (files[i].Name() == testFileName)
	}

	assert.True(t, foundFile)

	outTestDir := filepath.Join(outDir, "test")
	files, err = os.ReadDir(outTestDir)
	if err != nil {
		t.Fatalf("Failed to list files in dir %s: %v", outTestDir, err)
	}

	for i := range files {
		foundFile = foundFile || (files[i].Name() == testFileName)
	}

	assert.True(t, foundFile)
}

func setupFS(tb testing.TB) (mountDir, outDir string) {
	tb.Helper()

	wd, err := os.Getwd()
	if err != nil {
		tb.Fatalf("Failed to determine working directory: %v", err)
	}

	mountDir = tb.TempDir()
	outDir = tb.TempDir()

	spec := storage.Layers{
		{
			Target: "work",
			Source: wd,
		},
	}
	ufs := storage.NewUnionFS(outDir)
	srv, err := storage.MountUnionFS(ufs, mountDir, spec)
	if err != nil {
		tb.Fatalf("Failed to mount fs: %v", err)
	}

	go func() {
		srv.Wait()
	}()

	tb.Cleanup(func() {
		if err := srv.Unmount(); err != nil {
			tb.Errorf("Failed to unmount: %v", err)
		}
	})

	if err := srv.WaitMount(); err != nil {
		tb.Fatalf("Failed to wait for ready mount: %v", err)
	}

	return mountDir, outDir
}
