package modules

import (
	"encoding/json"
	"reflect"

	"github.com/docker/docker/api/types/mount"

	"github.com/hashicorp/hcl/v2"
	"github.com/hashicorp/hcl/v2/gohcl"

	commonv1 "code.icb4dc0.de/buildr/api/generated/common/v1"
	"code.icb4dc0.de/buildr/common/protocol"
)

var (
	_ ModuleWithMeta           = (*Metadata[Module])(nil)
	_ json.Unmarshaler         = (*Metadata[Module])(nil)
	_ protocol.SpecUnmarshaler = (*Metadata[Module])(nil)
)

type ContainerCapabilities struct {
	Add  []string `hcl:"add,optional"`
	Drop []string `hcl:"drop,optional"`
}

type ContainerBindMount struct {
	Target   string `hcl:",label"`
	Source   string `hcl:"source"`
	ReadOnly bool   `hcl:"read_only,optional"`
}

type ContainerTmpfsMount struct {
	Target   string `hcl:",label"`
	ReadOnly bool   `hcl:"read_only,optional"`
	Size     int64  `hcl:"size"`
}

type ContainerVolumeMount struct {
	Target   string `hcl:",label"`
	Name     string `hcl:"name,optional"`
	ReadOnly bool   `hcl:"read_only,optional"`
	NoCopy   bool   `hcl:"no_copy,optional"`
}

type ContainerSpec struct {
	Capabilities *ContainerCapabilities `hcl:"capabilities,block"`
	Image        string                 `hcl:"image"`
	User         string                 `hcl:"user,optional"`
	VolumeMounts []ContainerVolumeMount `hcl:"volume_mount,block"`
	TmpfsMounts  []ContainerTmpfsMount  `hcl:"tmpfs_mount,block"`
	BindMounts   []ContainerBindMount   `hcl:"bind_mount,block"`
	Privileged   bool                   `hcl:"privileged,optional"`
}

func (s ContainerSpec) Mounts() []mount.Mount {
	mounts := make([]mount.Mount, 0, len(s.VolumeMounts)+len(s.TmpfsMounts)+len(s.BindMounts))

	for i := range s.VolumeMounts {
		v := s.VolumeMounts[i]

		mounts = append(mounts, mount.Mount{
			Type:     mount.TypeVolume,
			Source:   v.Name,
			Target:   v.Target,
			ReadOnly: v.ReadOnly,
			VolumeOptions: &mount.VolumeOptions{
				NoCopy: v.NoCopy,
			},
		})
	}

	for i := range s.BindMounts {
		v := s.BindMounts[i]

		mounts = append(mounts, mount.Mount{
			Type:     mount.TypeBind,
			Source:   v.Source,
			Target:   v.Target,
			ReadOnly: v.ReadOnly,
		})
	}

	for i := range s.TmpfsMounts {
		v := s.TmpfsMounts[i]

		mounts = append(mounts, mount.Mount{
			Type:     mount.TypeTmpfs,
			Target:   v.Target,
			ReadOnly: v.ReadOnly,
			TmpfsOptions: &mount.TmpfsOptions{
				SizeBytes: v.Size,
			},
		})
	}

	return mounts
}

//nolint:stylecheck // can't rename Id due to conflict with interface
type Metadata[T Module] struct {
	Module       T                 `hcl:",remain"`
	InputMapping map[string]string `hcl:"input_mapping,optional"`
	Container    *ContainerSpec    `hcl:"container,block"`
	Skip         *bool             `hcl:"skip_when,optional"`
	Id           string            `hcl:"id" hcl_write:"-"`
	ModuleName   string            `hcl:"name" hcl_write:"-"`
	OutputDir    string            `hcl:"out_dir" hcl_write:"-"`
	Deps         []string          `hcl:"depends_on,optional"`
}

func (m *Metadata[T]) UnmarshalModuleSpec(val *commonv1.ModuleSpec) error {
	v := reflect.ValueOf(m.Module)
	if v.Kind() == reflect.Ptr {
		return protocol.Unmarshal(val, m.Module)
	}

	return protocol.Unmarshal(val, &m.Module)
}

func (m *Metadata[T]) SetName(name string) {
	m.ModuleName = name
}

func (m *Metadata[T]) SetModule(mod Module) {
	if t, ok := mod.(T); ok {
		m.Module = t
	}
}

func (m *Metadata[T]) ShouldSkip() bool {
	if m.Skip == nil {
		return false
	}

	return *m.Skip
}

func (m *Metadata[T]) UnmarshalJSON(bytes []byte) error {
	if reflect.ValueOf(m.Module).Kind() != reflect.Pointer {
		return json.Unmarshal(bytes, &m.Module)
	}
	return json.Unmarshal(bytes, m.Module)
}

func (m *Metadata[T]) UnmarshalHCL(body hcl.Body, ctx *hcl.EvalContext) hcl.Diagnostics {
	return gohcl.DecodeBody(body, ctx, m)
}

func (m *Metadata[T]) Execute(ctx ExecutionContext) error {
	return m.Module.Execute(ctx)
}

func (m *Metadata[T]) Category() Category {
	return m.Module.Category()
}

func (m *Metadata[T]) Type() string {
	return m.Module.Type()
}

func (m *Metadata[T]) Unwrap() Module {
	return m.Module
}

func (m *Metadata[T]) InputMappings() (mappings map[string]string) {
	mappings = m.InputMapping
	if mappings == nil {
		mappings = make(map[string]string)
	}

	return mappings
}

func (m *Metadata[T]) OutDir() string {
	return m.OutputDir
}

func (m *Metadata[T]) Dependencies() []string {
	return m.Deps
}

func (m *Metadata[T]) ID() string {
	return m.Id
}

func (m *Metadata[T]) Name() string {
	return m.ModuleName
}

func (m *Metadata[T]) ContainerSpec() *ContainerSpec {
	return m.Container
}
