package archive

import (
	"compress/gzip"
	"errors"
	"fmt"
	"os"
	"path/filepath"

	"code.icb4dc0.de/buildr/buildr/internal/ioutils"
	"code.icb4dc0.de/buildr/buildr/modules"
)

var (
	_ modules.Module = (*GzipFileCompression)(nil)
	_ modules.Helper = (*GzipFileCompression)(nil)
)

type GzipFileCompression struct {
	Source  string `hcl:"src"`
	OutFile string `hcl:"file_name,optional"`
}

func (GzipFileCompression) Category() modules.Category {
	return modules.CategoryPackage
}

func (GzipFileCompression) Type() string {
	return "gzip_file"
}

func (b GzipFileCompression) Execute(ctx modules.ExecutionContext) (err error) {
	if b.Source == "" {
		return fmt.Errorf("source file path may not be empty")
	}

	if !filepath.IsAbs(b.Source) {
		b.Source = filepath.Join(ctx.WorkingDir(), b.Source)
	}

	source, err := os.Open(b.Source)
	if err != nil {
		return fmt.Errorf("failed to open source file: %w", err)
	}

	defer func() {
		err = errors.Join(err, source.Close())
	}()

	if b.OutFile == "" {
		b.OutFile = filepath.Join(ctx.OutDir(), filepath.Base(b.Source)+".gz")
	} else if !filepath.IsAbs(b.OutFile) {
		b.OutFile = filepath.Join(ctx.OutDir(), b.OutFile)
	}

	outFile, err := os.Create(b.OutFile)
	if err != nil {
		return fmt.Errorf("failed to create output file: %w", err)
	}

	defer func() {
		err = errors.Join(err, outFile.Close())
	}()

	writer, err := gzip.NewWriterLevel(outFile, gzip.DefaultCompression)
	if err != nil {
		return fmt.Errorf("failed to create gzip writer: %w", err)
	}
	defer func() {
		err = errors.Join(err, writer.Close())
	}()

	_, err = ioutils.CopyWithPooledBuffer(writer, source)

	return err
}
