package archive_test

import (
	"archive/tar"
	"errors"
	"fmt"
	"io"
	"math/rand"
	"net/http"
	"os"
	"strings"
	"testing"
	"time"

	"github.com/klauspost/compress/zip"
	"github.com/klauspost/pgzip"
	"github.com/stretchr/testify/assert"
)

var replacer = strings.NewReplacer("_", "", "-", "", "/", "")

func createSampleFileAt(tb testing.TB, dir string, size uint64) string {
	tb.Helper()

	//nolint:gosec // only used for random file creation
	random := rand.New(rand.NewSource(time.Now().Unix()))
	f, err := os.CreateTemp(dir, fmt.Sprintf("%s-*", replacer.Replace(tb.Name())))
	if err != nil {
		tb.Fatalf("failed to create temp file: %v", err)
	}
	defer func() {
		if err := f.Close(); err != nil {
			tb.Fatalf("failed to close temp file: %v", err)
		}
	}()

	buf := make([]byte, 1024)
	n, err := random.Read(buf)
	if err != nil {
		tb.Fatalf("failed to read random bytes: %v", err)
	}

	for size > 0 {
		if size < 1024 {
			buf = buf[:size]
		}

		if _, err := f.Write(buf[:n]); err != nil {
			tb.Fatalf("failed to write random bytes to temp file: %v", err)
		}

		size -= uint64(len(buf))
	}

	return f.Name()
}

func createSampleFile(tb testing.TB, size uint64) string {
	tb.Helper()

	return createSampleFileAt(tb, tb.TempDir(), size)
}

func listFilesInArchive(tb testing.TB, filePath string) (fileList []string) {
	tb.Helper()

	info, err := os.Stat(filePath)
	if err != nil {
		tb.Fatalf("failed to stat file: %v", err)
	}

	f, err := os.Open(filePath)
	if err != nil {
		tb.Fatalf("failed to open tar file: %v", err)
	}

	peekBuf := make([]byte, 512)
	read, err := f.ReadAt(peekBuf, 0)
	if err != nil && !errors.Is(err, io.EOF) {
		tb.Fatalf("failed to read tar header: %v", err)
	}

	switch http.DetectContentType(peekBuf[:read]) {
	case "application/zip":
		return listFilesInZip(tb, f, info.Size())
	case "application/x-gzip":
		return listFilesInTarGzip(tb, f)
	case "application/octet-stream":
		return listFilesInTar(tb, f)
	default:
		tb.Fatal("unsupported archive type")
		return nil
	}
}

func listFilesInZip(tb testing.TB, reader io.ReaderAt, size int64) (fileList []string) {
	tb.Helper()

	f, err := zip.NewReader(reader, size)
	if !assert.NoError(tb, err, "failed to open zip file") {
		tb.FailNow()
	}

	for _, file := range f.File {
		fileList = append(fileList, file.Name)
	}

	return fileList
}

func listFilesInTarGzip(tb testing.TB, reader io.Reader) (fileList []string) {
	tb.Helper()

	gzipReader, err := pgzip.NewReader(reader)
	if !assert.NoError(tb, err, "failed to create gzip reader") {
		tb.FailNow()
	}

	return listFilesInTar(tb, gzipReader)
}

func listFilesInTar(tb testing.TB, reader io.Reader) (fileList []string) {
	tb.Helper()
	tarReader := tar.NewReader(reader)

	for header, err := tarReader.Next(); err == nil; header, err = tarReader.Next() {
		fileList = append(fileList, header.Name)
	}

	return fileList
}
