package archive

import (
	"errors"
	"fmt"
	"io"
	"log/slog"
	"os"
	"path/filepath"

	"github.com/klauspost/pgzip"

	"code.icb4dc0.de/buildr/buildr/internal/archive"
	"code.icb4dc0.de/buildr/buildr/internal/ignore"
	"code.icb4dc0.de/buildr/buildr/internal/ioutils"
	"code.icb4dc0.de/buildr/buildr/modules"
)

var (
	_ modules.Module = (*TarArchive)(nil)
	_ modules.Helper = (*TarArchive)(nil)
)

type CompressionType string

func (t CompressionType) Writer(outer io.WriteCloser) (io.WriteCloser, error) {
	switch t {
	case CompressionNone:
		return ioutils.NoOpWriteCloser(outer), nil
	case CompressionGzip:
		return pgzip.NewWriter(outer), nil
	default:
		return nil, fmt.Errorf("unknown compression type: %s", t)
	}
}

const (
	CompressionNone CompressionType = ""
	CompressionGzip CompressionType = "gzip"
)

type TarArchive struct {
	Name        string          `hcl:"archive_name"`
	Compression CompressionType `hcl:"compression,optional"`
	Content     []string        `hcl:"content,optional"`
}

func (TarArchive) Category() modules.Category {
	return modules.CategoryPackage
}

func (TarArchive) Type() string {
	return "tar_archive"
}

func (t TarArchive) Execute(ctx modules.ExecutionContext) (err error) {
	if t.Name == "" {
		return ErrEmptyArchiveName
	}

	logger := ctx.Logger()
	logger.Info("Packaging archive", slog.String("archive_name", t.Name))

	ignorer, err := ignore.NewIgnorer(ctx.WorkingDir())
	if err != nil {
		return err
	}

	ta := archive.Tar{
		Ignorer: ignorer,
	}

	if t.Content == nil {
		if err := t.archiveWorkingDir(&ta, ctx.WorkingDir()); err != nil {
			return err
		}
	} else {
		if err := t.archiveContent(&ta, ctx.WorkingDir(), logger); err != nil {
			return err
		}
	}

	outFilePath := t.Name
	if !filepath.IsAbs(t.Name) {
		outFilePath = filepath.Join(ctx.OutDir(), t.Name)
	}

	logger.Debug("Creating out file", slog.String("out_file", outFilePath))

	outFile, err := os.Create(outFilePath)
	if err != nil {
		return fmt.Errorf("failed to create out file for tar archive: %w", err)
	}

	defer func() {
		err = errors.Join(err, outFile.Close())
	}()

	compressedWriter, err := t.Compression.Writer(outFile)
	if err != nil {
		return err
	}

	defer func() {
		err = errors.Join(err, compressedWriter.Close())
	}()

	return ta.Write(compressedWriter)
}

func (TarArchive) archiveWorkingDir(ta *archive.Tar, workingDir string) error {
	return ta.Add(workingDir, "")
}

func (t TarArchive) archiveContent(ta *archive.Tar, workingDir string, logger *slog.Logger) error {
	for _, c := range t.Content {
		if !filepath.IsAbs(c) {
			c = filepath.Join(workingDir, c)
		}

		logger.Debug("Adding file to tar archive", slog.String("file", c))

		if err := ta.Add(c, ""); err != nil {
			return err
		}
	}

	return nil
}
