package archive_test

import (
	"context"
	"path/filepath"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"code.icb4dc0.de/buildr/buildr/modules"
	"code.icb4dc0.de/buildr/buildr/modules/packaging/archive"
)

func TestTarArchive_Execute_SingleFileNoContent(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name        string
		archiveName string
		compression archive.CompressionType
	}{
		{
			name:        "Single file - without compression",
			archiveName: "out.tar",
		},
		{
			name:        "Single file - with GZIP compression",
			archiveName: "out.tar.gz",
			compression: archive.CompressionGzip,
		},
	}

	for _, tt := range tests {
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			var (
				workingDir = t.TempDir()
				outDir     = t.TempDir()
			)

			createSampleFileAt(t, workingDir, 500*1024)

			ta := archive.TarArchive{
				Name:        tt.archiveName,
				Compression: tt.compression,
			}

			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Second)
			t.Cleanup(cancel)

			execCtx := modules.TestExecutionContext{
				Context:         ctx,
				TB:              t,
				OutputDirectory: outDir,
				TestWorkingDir:  workingDir,
			}

			if !assert.NoError(t, ta.Execute(execCtx), "failed to execute tar module") {
				t.FailNow()
			}

			fileList := listFilesInArchive(t, filepath.Join(outDir, tt.archiveName))
			assert.Len(t, fileList, 1)
		})
	}
}

func TestTarArchive_Execute_MultipleFilesWithContent(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                 string
		archiveName          string
		compression          archive.CompressionType
		numberOfFiles        int
		numberOfContentFiles int
		fileSize             uint64
	}{
		{
			name:                 "Single file and content - plain tar archive",
			archiveName:          "out.tar",
			numberOfFiles:        1,
			numberOfContentFiles: 1,
			fileSize:             100 * 1024,
		},
		{
			name:                 "Single file and content - tar archive with gzip compression",
			archiveName:          "out.tar.gz",
			compression:          archive.CompressionGzip,
			numberOfFiles:        1,
			numberOfContentFiles: 1,
			fileSize:             100 * 1024,
		},
		{
			name:                 "Multiple files - sub-selection in content - plain tar",
			archiveName:          "out.tar",
			numberOfFiles:        5,
			numberOfContentFiles: 3,
			fileSize:             100 * 1024,
		},
		{
			name:                 "Multiple files - sub-selection in content - tar archive with gzip compression",
			archiveName:          "out.tar.gz",
			compression:          archive.CompressionGzip,
			numberOfFiles:        5,
			numberOfContentFiles: 3,
			fileSize:             100 * 1024,
		},
	}

	for _, tt := range tests {
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			workingDir := t.TempDir()
			content := make([]string, 0, tt.numberOfContentFiles)

			for i, j := 0, 0; i < 5; i, j = i+1, j+1 {
				f := createSampleFileAt(t, workingDir, 100*1024)
				if j < tt.numberOfContentFiles {
					relFilePath, err := filepath.Rel(workingDir, f)
					if !assert.NoError(t, err, "failed to get relative path") {
						t.FailNow()
					}
					content = append(content, relFilePath)
				}
			}

			ta := archive.TarArchive{
				Name:        tt.archiveName,
				Compression: tt.compression,
				Content:     content,
			}

			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Second)
			t.Cleanup(cancel)

			outDir := t.TempDir()

			execCtx := modules.TestExecutionContext{
				Context:         ctx,
				TB:              t,
				OutputDirectory: outDir,
				TestWorkingDir:  workingDir,
			}

			if !assert.NoError(t, ta.Execute(execCtx), "failed to execute tar module") {
				t.FailNow()
			}
		})
	}
}
