package archive

import (
	"errors"
	"fmt"
	"io/fs"
	"log/slog"
	"os"
	"path"
	"path/filepath"

	"code.icb4dc0.de/buildr/buildr/internal/ioutils"
	"code.icb4dc0.de/buildr/buildr/modules"

	"github.com/klauspost/compress/zip"
)

var (
	_ modules.Module = (*ZipArchive)(nil)
	_ modules.Helper = (*ZipArchive)(nil)
)

type ZipArchive struct {
	Name      string   `hcl:"archive_name"`
	WrapInDir string   `hcl:"wrap_in_directory,optional"`
	Content   []string `hcl:"content,optional"`
}

func (ZipArchive) Category() modules.Category {
	return modules.CategoryPackage
}

func (ZipArchive) Type() string {
	return "zip_archive"
}

func (z ZipArchive) Execute(ctx modules.ExecutionContext) (err error) {
	if z.Name == "" {
		return ErrEmptyArchiveName
	}

	logger := ctx.Logger()
	logger.Info("Packaging archive", slog.String("archive_name", z.Name))

	outFilePath := z.Name
	if !filepath.IsAbs(z.Name) {
		outFilePath = filepath.Join(ctx.OutDir(), z.Name)
	}

	logger.Debug("Creating out file", slog.String("out_file", outFilePath))
	zipFile, err := os.Create(outFilePath)
	if err != nil {
		return fmt.Errorf("failed to create out file for zip archive: %w", err)
	}
	defer func() {
		err = errors.Join(err, zipFile.Close())
	}()

	zipWriter := zip.NewWriter(zipFile)

	defer func() {
		if zipCloseErr := zipWriter.Close(); zipCloseErr != nil {
			err = errors.Join(err, fmt.Errorf("failed to close zip archive: %w", zipCloseErr))
		}
	}()

	if z.Content == nil {
		return z.compressAllInWorkingDir(logger, zipWriter, ctx.WorkingDir())
	}

	return z.compressContent(logger, zipWriter, ctx.WorkingDir())
}

func (z ZipArchive) compressAllInWorkingDir(logger *slog.Logger, zipWriter *zip.Writer, workingDir string) error {
	if err := filepath.WalkDir(workingDir, z.walker(logger, zipWriter, workingDir, z.WrapInDir)); err != nil {
		return fmt.Errorf("failed to recursively add files to zip archive: %w", err)
	}

	return nil
}

func (z ZipArchive) compressContent(logger *slog.Logger, zipWriter *zip.Writer, workingDir string) error {
	for _, srcPath := range z.Content {
		if !filepath.IsAbs(srcPath) {
			srcPath = filepath.Join(workingDir, srcPath)
		}

		info, err := os.Stat(srcPath)
		if err != nil {
			return fmt.Errorf("failed to get file info for file %s: %w", srcPath, err)
		}

		logger.Debug("Adding file to zip archive", slog.String("file", srcPath))

		if err := addFileToArchive(zipWriter, srcPath, workingDir, "", info); err != nil {
			return fmt.Errorf("failed to add file %s to zip archive: %w", srcPath, err)
		}
	}

	return nil
}

func (z ZipArchive) walker(logger *slog.Logger, zipWriter *zip.Writer, basePath, commonPrefix string) fs.WalkDirFunc {
	return func(currentPath string, d fs.DirEntry, err error) error {
		if err != nil {
			return err
		}

		if d.IsDir() {
			return nil
		}

		logger.Debug("Adding file to zip archive", slog.String("file", currentPath))

		info, err := d.Info()
		if err != nil {
			return fmt.Errorf("failed to get file info for file %s: %w", currentPath, err)
		}

		return addFileToArchive(zipWriter, currentPath, basePath, commonPrefix, info)
	}
}

func addFileToArchive(zipWriter *zip.Writer, filePath, basePath, commonPrefix string, info fs.FileInfo) (err error) {
	archivePath := filepath.ToSlash(filePath)
	if filepath.IsAbs(archivePath) {
		if archivePath, err = filepath.Rel(basePath, archivePath); err != nil {
			return err
		}
	}

	if commonPrefix != "" {
		archivePath = path.Join(commonPrefix, archivePath)
	}

	header, err := zip.FileInfoHeader(info)
	if err != nil {
		return fmt.Errorf("failed to get file header for file %s: %w", filePath, err)
	}

	sourceFile, err := os.Open(filePath)
	if err != nil {
		return fmt.Errorf("failed to open file %s: %w", filePath, err)
	}

	defer sourceFile.Close()

	header.Method = zip.Deflate
	header.Name = filepath.ToSlash(archivePath)

	zipFileWriter, err := zipWriter.CreateHeader(header)
	if err != nil {
		return fmt.Errorf("failed to write header for file %s: %w", filePath, err)
	}

	_, err = ioutils.CopyWithPooledBuffer(zipFileWriter, sourceFile)
	if err != nil {
		return fmt.Errorf("failed to copy file to zip archive %s: %w", filePath, err)
	}

	return nil
}
