package archive_test

import (
	"context"
	"path/filepath"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"code.icb4dc0.de/buildr/buildr/modules"
	"code.icb4dc0.de/buildr/buildr/modules/packaging/archive"
)

func TestZipArchive_Execute_SingleFileNoContent(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name      string
		wrapInDir string
	}{
		{
			name: "Single file - without wrapping in directory",
		},
		{
			name:      "Single file - with wrapping in directory",
			wrapInDir: "out",
		},
	}

	for _, tt := range tests {
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			var (
				workingDir = t.TempDir()
				outDir     = t.TempDir()
			)

			createSampleFileAt(t, workingDir, 500*1024)

			zip := archive.ZipArchive{
				Name:      "out.zip",
				WrapInDir: tt.wrapInDir,
			}

			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Second)
			t.Cleanup(cancel)

			execCtx := modules.TestExecutionContext{
				Context:         ctx,
				TB:              t,
				OutputDirectory: outDir,
				TestWorkingDir:  workingDir,
			}

			if !assert.NoError(t, zip.Execute(execCtx), "failed to execute zip module") {
				t.FailNow()
			}

			outfilePath := filepath.Join(outDir, "out.zip")
			assert.FileExists(t, outfilePath)
			assert.Len(t, listFilesInArchive(t, outfilePath), 1)
		})
	}
}

func TestZipArchive_Execute_MultipleFilesNoContent(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name          string
		wrapInDir     string
		numberOfFiles int
		fileSize      uint64
	}{
		{
			name:          "Multiple files - without wrapping in directory",
			numberOfFiles: 5,
			fileSize:      100 * 1024,
		},
		{
			name:          "Multiple files - with wrapping in directory",
			wrapInDir:     "out",
			numberOfFiles: 5,
			fileSize:      100 * 1024,
		},
	}

	for _, tt := range tests {
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			workingDir := t.TempDir()

			for i := 0; i < 5; i++ {
				createSampleFileAt(t, workingDir, 100*1024)
			}

			zip := archive.ZipArchive{Name: "out.zip"}

			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Second)
			t.Cleanup(cancel)

			outDir := t.TempDir()

			execCtx := modules.TestExecutionContext{
				Context:         ctx,
				TB:              t,
				OutputDirectory: outDir,
				TestWorkingDir:  workingDir,
			}

			if !assert.NoError(t, zip.Execute(execCtx), "failed to execute zip module") {
				t.FailNow()
			}

			outFilePath := filepath.Join(outDir, "out.zip")
			assert.FileExists(t, outFilePath)
			assert.Len(t, listFilesInArchive(t, outFilePath), tt.numberOfFiles)
		})
	}
}

func TestZipArchive_Execute_MultipleFilesWithContent(t *testing.T) {
	t.Parallel()

	tests := []struct {
		name                 string
		wrapInDir            string
		numberOfFiles        int
		numberOfContentFiles int
		fileSize             uint64
	}{
		{
			name:                 "single file - single content",
			numberOfFiles:        1,
			numberOfContentFiles: 1,
			fileSize:             100 * 1024,
		},
		{
			name:                 "Multiple files - sub-selection in content",
			wrapInDir:            "out",
			numberOfFiles:        5,
			numberOfContentFiles: 3,
			fileSize:             100 * 1024,
		},
	}

	for _, tt := range tests {
		tt := tt
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			workingDir := t.TempDir()
			content := make([]string, 0, tt.numberOfContentFiles)

			for i, j := 0, 0; i < 5; i, j = i+1, j+1 {
				f := createSampleFileAt(t, workingDir, 100*1024)
				if j < tt.numberOfContentFiles {
					relFilePath, err := filepath.Rel(workingDir, f)
					if !assert.NoError(t, err, "failed to get relative path") {
						t.FailNow()
					}
					content = append(content, relFilePath)
				}
			}

			zip := archive.ZipArchive{
				Name:    "out.zip",
				Content: content,
			}

			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Second)
			t.Cleanup(cancel)

			outDir := t.TempDir()

			execCtx := modules.TestExecutionContext{
				Context:         ctx,
				TB:              t,
				OutputDirectory: outDir,
				TestWorkingDir:  workingDir,
			}

			if !assert.NoError(t, zip.Execute(execCtx), "failed to execute zip module") {
				t.FailNow()
			}

			outFilePath := filepath.Join(outDir, "out.zip")
			assert.FileExists(t, outFilePath)
			assert.Len(t, listFilesInArchive(t, outFilePath), tt.numberOfContentFiles)
		})
	}
}
