package ociimg

import (
	"context"
	"errors"
	"fmt"
	"log/slog"

	"github.com/google/go-containerregistry/pkg/name"
	v1 "github.com/google/go-containerregistry/pkg/v1"
	"github.com/google/go-containerregistry/pkg/v1/daemon"
)

func newContainerDaemon(imageName string, logger *slog.Logger, tags []string, cli daemon.Client) containerDaemon {
	return containerDaemon{
		imageName: imageName,
		logger:    logger,
		client:    cli,
		tags:      tags,
	}
}

type containerDaemon struct {
	imageName string
	logger    *slog.Logger
	client    daemon.Client
	tags      []string
}

func (d containerDaemon) Publish(ctx context.Context, result imageOrIndex) (name.Reference, error) {
	if len(d.tags) == 0 {
		return nil, errors.New("at least one tag must be specified")
	}

	var img v1.Image
	switch i := result.(type) {
	case v1.Image:
		img = i
	case v1.ImageIndex:
		im, err := i.IndexManifest()
		if err != nil {
			return nil, err
		}

		if len(im.Manifests) < 1 {
			return nil, errors.New("index is empty")
		}

		img, err = i.Image(im.Manifests[0].Digest)
		if err != nil {
			return nil, err
		}

	default:
		return nil, fmt.Errorf("failed to interpret %s result as image: %v", d.imageName, result)
	}

	defaultTag, err := name.NewTag(fmt.Sprintf("%s:%s", d.imageName, d.tags[0]))
	if err != nil {
		return nil, err
	}

	d.logger.Info("Loading", slog.Any("defaultTag", defaultTag))

	if resp, err := daemon.Write(defaultTag, img, d.getOpts(ctx)...); err != nil {
		d.logger.Debug("daemon.Write", slog.String("response", resp))
		return nil, err
	}

	d.logger.Info("Loaded", slog.Any("defaultTag", defaultTag))

	for _, tagName := range d.tags[1:] {
		d.logger.Info("Adding tag", slog.String("tag", tagName))

		tag, err := name.NewTag(fmt.Sprintf("%s:%s", d.imageName, tagName))
		if err != nil {
			return nil, err
		}

		if err := daemon.Tag(defaultTag, tag, d.getOpts(ctx)...); err != nil {
			return nil, err
		}
		d.logger.Info("Added tag", slog.String("tag", tagName))
	}

	return &defaultTag, nil
}

func (d containerDaemon) getOpts(ctx context.Context) []daemon.Option {
	return []daemon.Option{
		daemon.WithContext(ctx),
		daemon.WithClient(d.client),
	}
}
