package plugin

import (
	"context"
	"errors"
	"fmt"

	"connectrpc.com/connect"
	"github.com/tetratelabs/wazero"
	"github.com/tetratelabs/wazero/imports/wasi_snapshot_preview1"

	commonv1 "code.icb4dc0.de/buildr/api/generated/common/v1"
	wasiv1 "code.icb4dc0.de/buildr/api/generated/wasi/v1"
	"code.icb4dc0.de/buildr/api/generated/wasi/v1/rpcv1connect"
	"code.icb4dc0.de/buildr/common/wasirpc"
)

func (m Module) BinaryName(ctx context.Context) (binaryName string, err error) {
	runtimeConfig := wazero.
		NewRuntimeConfig().
		WithCloseOnContextDone(true)

	r := wazero.NewRuntimeWithConfig(ctx, runtimeConfig)
	defer func() {
		closeCtx, cancel := context.WithTimeout(context.Background(), runtimeClosingTimeout)
		defer cancel()
		err = errors.Join(err, r.Close(closeCtx))
	}()

	_, err = r.NewHostModuleBuilder("buildr").
		NewFunctionBuilder().WithFunc(dummyWithResult).Export("log_msg").
		NewFunctionBuilder().WithFunc(dummyWithResult).Export("get_state").
		NewFunctionBuilder().WithFunc(dummyWithResult).Export("set_state").
		NewFunctionBuilder().WithFunc(dummyWithResult).Export("exec").
		NewFunctionBuilder().WithFunc(dummyWithResult).Export("lookPath").
		Instantiate(ctx)

	if err != nil {
		return "", fmt.Errorf("failed to instantiate host module builder: %w", err)
	}

	closer, err := wasi_snapshot_preview1.Instantiate(ctx, r)
	if err != nil {
		return "", fmt.Errorf("failed to instantiate wasi snapshot preview1: %w", err)
	}

	defer func() {
		err = errors.Join(err, closer.Close(context.Background()))
	}()

	pluginPayload, err := m.PluginPayload.Bytes()
	if err != nil {
		return "", err
	}

	mod, err := r.Instantiate(ctx, pluginPayload)
	if err != nil {
		return "", fmt.Errorf("failed to instantiate WASI plugin: %w", err)
	}

	moduleClient, err := wasirpc.NewModuleClient(mod)
	if err != nil {
		return "", fmt.Errorf("failed to create module client: %w", err)
	}

	wasiClient := rpcv1connect.NewWasiExecutorServiceClient(moduleClient, fmt.Sprintf("wasi://%s", m.PluginType))

	req := &wasiv1.BinaryNameRequest{
		ModuleReference: &commonv1.ModuleReference{
			ModuleCategory: m.PluginCategory,
			ModuleType:     m.PluginType,
		},
		Spec: m.modSpec,
	}

	resp, err := wasiClient.BinaryName(ctx, connect.NewRequest(req))
	if err != nil {
		return "", fmt.Errorf("failed to get plugin binary name: %w", err)
	}

	return resp.Msg.Name, nil
}
