package modules

import (
	"errors"
	"fmt"
	"sync"

	"github.com/hashicorp/hcl/v2"
)

var ErrNoSuchModule = errors.New("no such module registered")

type Registration interface {
	RegisterAt(registry *TypeRegistry)
}

type RegistrationFunc func(registry *TypeRegistry)

func (f RegistrationFunc) RegisterAt(registry *TypeRegistry) {
	f(registry)
}

func NewRegistry(registrations ...Registration) *TypeRegistry {
	r := &TypeRegistry{
		registrations: make(map[moduleSpec]Factory),
	}

	for i := range registrations {
		registrations[i].RegisterAt(r)
	}

	return r
}

type TypeRegistry struct {
	registrations map[moduleSpec]Factory
	lock          sync.RWMutex
}

func (r *TypeRegistry) Inventory() (inventory map[Category][]string) {
	r.lock.RLock()
	defer r.lock.RUnlock()

	inventory = make(map[Category][]string, len(r.registrations))

	for spec := range r.registrations {
		mods := inventory[spec.TypeName]
		inventory[spec.TypeName] = append(mods, spec.ModuleName)
	}

	return inventory
}

func (r *TypeRegistry) Create(moduleCategory Category, moduleType string) (ModuleWithMeta, error) {
	r.lock.RLock()
	defer r.lock.RUnlock()

	s := specOf(moduleCategory, moduleType)
	if m, ok := r.registrations[s]; ok {
		return m.Create(), nil
	} else {
		return nil, fmt.Errorf("%w: %s/%s", ErrNoSuchModule, moduleCategory, moduleType)
	}
}

func (r *TypeRegistry) CreateFromHCL(
	moduleCategory Category,
	moduleType string,
	body hcl.Body,
	hclCtx *hcl.EvalContext,
	diagsWriter hcl.DiagnosticWriter,
) (ModuleWithMeta, error) {
	mod, err := r.Create(moduleCategory, moduleType)
	if err != nil {
		return nil, err
	}

	if diags := mod.UnmarshalHCL(body, hclCtx); diags.HasErrors() {
		_ = diagsWriter.WriteDiagnostics(diags)
		return nil, diags
	}

	unwrapped := mod.Unwrap()
	if initializer, ok := unwrapped.(Initializer); ok {
		if initialized, err := initializer.Init(hclCtx); err != nil {
			return nil, err
		} else {
			mod.SetModule(initialized)
		}
	}

	return mod, nil
}

func (r *TypeRegistry) RegisterModule(factory Factory) {
	r.lock.Lock()
	defer r.lock.Unlock()
	m := factory.Create()
	s := specOf(m.Category(), m.Type())
	if _, ok := r.registrations[s]; !ok {
		r.registrations[s] = factory
	}
}
